# ------------------------------------------------------------------------------
#
# Project: pygml <https://github.com/geopython/pygml>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from typing import Optional
from collections.abc import Sequence

from .types import GeomDict


def get_dimensionality(geometry: GeomDict) -> Optional[int]:
    """ Returns the dimensionality of a given GeoJSON geometry.
        This is obtained by descending into the first coordinate
        and using its length.
        When no coordinates can be retrieved (e.g: in case of
        GeometryCollections) None is returned.

        >>> get_dimensionality({
        ...     'type': 'Polygon',
        ...     'coordinates': [
        ...         [
        ...             (0.5, 1.0),
        ...             (0.5, 2.0),
        ...             (1.5, 2.0),
        ...             (1.5, 1.0),
        ...             (0.5, 1.0)
        ...         ]
        ...     ]
        ... })
        2
        >>> get_dimensionality({
        ...     'type': 'MultiPoint',
        ...     'coordinates': [
        ...         (1.0, 1.0, 1.0),
        ...         (2.0, 2.0, 1.0),
        ...     ]
        ... })
        3
        >>> get_dimensionality({
        ...     'type': 'GeometryCollection',
        ...     'geometries': [
        ...         {
        ...             'type': 'Point',
        ...             'coordinates': (1.0, 1.0)
        ...         },
        ...         {
        ...             'type': 'Polygon',
        ...             'coordinates': [
        ...                 [(1.0, 1.0)],
        ...                 [(1.0, 1.0)],
        ...             ]
        ...         },
        ...     ]
        ... })
    """

    coordinates = geometry.get('coordinates')

    if coordinates:
        # drill down into nested coordinates
        while isinstance(coordinates[0], Sequence):
            coordinates = coordinates[0]
        return len(coordinates)
    return None



# ------------------------------------------------------------------------------
#
# Project: pygml <https://github.com/geopython/pygml>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from pygml.axisorder import get_crs_code
from typing import Callable, List

from lxml import etree
from lxml.builder import ElementMaker

from .basics import (
    parse_pos, parse_poslist, swap_coordinate_xy, swap_coordinates_xy
)
from .dimensionality import get_dimensionality
from .types import GeomDict
from .pre_v32 import (
    NAMESPACE as NAMESPACE_PRE32, parse_pre_v32, encode_pre_v32
)
from .v32 import NAMESPACE as NAMESPACE_32, parse_v32
from .v33 import NAMESPACE as NAMESPACE_33_CE, parse_v33_ce


NAMESPACE = 'http://www.georss.org/georss'
NSMAP = {'georss': NAMESPACE}


Element = etree._Element
Elements = List[Element]


def parse_georss(element: Element) -> GeomDict:
    """ Parses the GeoRSS basic elements to their respective GeoJSON
        representation. As all coordinates in GeoRSS are expressed in
        WGS84 and in Latitude/Longitude order, the coordinates are
        swapped to XY order.

        In case of georss:where, it is expected that it contains a
        single GML element which is parsed as either GML 3.1.1, GML 3.2
        or GML 3.3 CE.
    """
    qname = etree.QName(element.tag)
    if qname.namespace != NAMESPACE:
        raise ValueError(f'Unsupported namespace {qname.namespace}')

    bbox = None
    localname = qname.localname

    if localname == 'point':
        type_ = 'Point'
        coordinates = swap_coordinate_xy(parse_pos(element.text))
    elif localname == 'line':
        type_ = 'LineString'
        coordinates = swap_coordinates_xy(parse_poslist(element.text))
    elif localname == 'box':
        # boxes are expanded to Polygons, but store the 'bbox' value
        type_ = 'Polygon'
        low, high = swap_coordinates_xy(parse_poslist(element.text))
        lx, ly = low
        hx, hy = high
        coordinates = [
            [
                (lx, ly),
                (lx, hy),
                (hx, hy),
                (hx, ly),
                (lx, ly),
            ]
        ]
        bbox = (lx, ly, hx, hy)
    elif localname == 'polygon':
        type_ = 'Polygon'
        coordinates = [swap_coordinates_xy(parse_poslist(element.text))]

    elif localname == 'where':
        # special handling here: defer to the gml definition. Although,
        # only GML 3.1.1 is officially supported, we also allow GML 3.2 and 3.3
        if not len(element) == 1:
            raise ValueError(
                'Invalid number of child elements in georss:where'
            )
        child = element[0]
        child_namespace = etree.QName(child.tag).namespace

        if child_namespace == NAMESPACE_PRE32:
            return parse_pre_v32(child)
        elif child_namespace == NAMESPACE_32:
            return parse_v32(child)
        elif child_namespace == NAMESPACE_33_CE:
            return parse_v33_ce(child)
        else:
            raise ValueError(
                f'Unsupported child element in georss:where: {child.tag}'
            )
    else:
        raise ValueError(f'Unsupported georss element: {localname}')

    result = {
        'type': type_,
        'coordinates': coordinates,
    }
    if bbox:
        result['bbox'] = bbox

    return result


GEORSS = ElementMaker(namespace=NAMESPACE, nsmap=NSMAP)

GmlEncoder = Callable[[GeomDict, str], Element]


def encode_georss(geometry: GeomDict,
                  gml_encoder: GmlEncoder = encode_pre_v32) -> Element:
    """ Encodes a GeoJSON geometry as a GeoRSS ``lxml.etree.Element``.
        Tries to use the native GeoRSS elements ``point``, ``line``,
        or ``polygon`` when possible. Falls back to ``georss:where``
        with using the ``gml_encoder`` function (defaulting to GML 3.2):

            - MultiPoint, MultiLineString, MultiPolygon geometries
            - Polygons with interiors
            - GeometryCollections
            - any geometry with CRS other than CRS84 or EPSG:4326
            - when dealing with >2D geometries
    """
    type_ = geometry['type']
    coordinates = geometry.get('coordinates')
    crs = geometry.get('crs')
    dims = get_dimensionality(geometry)

    code = None
    if crs:
        crs_name = crs.get('properties', {}).get('name')
        code = get_crs_code(crs_name)

    if code in (None, 4326, 'CRS84') and dims == 2:
        if type_ == 'Point':
            return GEORSS(
                'point',
                ' '.join(
                    str(v) for v in swap_coordinate_xy(coordinates)
                )
            )

        elif type_ == 'LineString':
            return GEORSS(
                'line',
                ' '.join(
                    ' '.join(
                        str(v) for v in coordinate
                    ) for coordinate in swap_coordinates_xy(coordinates)
                )
            )

        elif type_ == 'Polygon':
            # only exterior
            if len(coordinates) == 1:
                return GEORSS(
                    'polygon',
                    ' '.join(
                        ' '.join(
                            str(v) for v in coordinate
                        ) for coordinate in swap_coordinates_xy(coordinates[0])
                    )
                )

    # fall back to GML encoding when we have:
    #   - MultiPoint, MultiLineString, MultiPolygon geometries
    #   - Polygons with interiors
    #   - GeometryCollections
    #   - any geometry with CRS other than CRS84 or EPSG4326
    #   - when dealing with >2D geometries
    return GEORSS(
        'where',
        gml_encoder(geometry, 'ID')
    )



# ------------------------------------------------------------------------------
#
# Project: pygml <https://github.com/geopython/pygml>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


from typing import Callable

from .types import Coordinates, Coordinate


def _make_number_parser(decimal: str) -> Callable[[str], float]:
    """ Helper to create a number parser with a potentially custom
        decimal separator. When this is not the '.' character, each
        number will replace the given decimal separator with '.'
        before calling the built-in `float` function.
    """
    if decimal == '.':
        return float

    def inner(value: str) -> float:
        return float(value.replace(decimal, '.'))

    return inner


def parse_coordinates(value: str, cs: str = ',', ts: str = ' ',
                      decimal: str = '.') -> Coordinates:
    """ Parses the the values of a gml:coordinates node to a list of
        lists of floats. Takes the coordinate separator and tuple
        separator into account, and also custom decimal separators.

        >>> parse_coordinates('12.34 56.7,89.10 11.12')
        [(12.34, 56.7), (89.1, 11.12)]
        >>> parse_coordinates('12.34 56.7;89.10 11.12', cs=';')
        [(12.34, 56.7), (89.1, 11.12)]
        >>> parse_coordinates('12.34:56.7,89.10:11.12', ts=':')
        [(12.34, 56.7), (89.1, 11.12)]
        >>> parse_coordinates('12.34:56.7;89.10:11.12', cs=';', ts=':')
        [(12.34, 56.7), (89.1, 11.12)]
        >>> parse_coordinates(
        ...     '12,34:56,7;89,10:11,12', cs=';', ts=':', decimal=','
        ... )
        [(12.34, 56.7), (89.1, 11.12)]
    """

    number_parser = _make_number_parser(decimal)

    return [
        tuple(
            number_parser(number)
            for number in coordinate.strip().split(ts)
        )
        for coordinate in value.strip().split(cs)
    ]


def parse_poslist(value: str, dimensions: int = 2) -> Coordinates:
    """ Parses the value of a single gml:posList to a `Coordinates`
        structure.

        >>> parse_poslist('12.34 56.7 89.10 11.12')
        [(12.34, 56.7), (89.1, 11.12)]
        >>> parse_poslist('12.34 56.7 89.10 11.12 13.14 15.16', dimensions=3)
        [(12.34, 56.7, 89.1), (11.12, 13.14, 15.16)]
        >>> parse_poslist('12.34 56.7 89.10 11.12', dimensions=3)
        Traceback (most recent call last):
            ...
        ValueError: Invalid dimensionality of pos list
    """
    raw = [float(v) for v in value.split()]
    if len(raw) % dimensions > 0:
        raise ValueError('Invalid dimensionality of pos list')

    return [
        tuple(raw[i:i + dimensions])
        for i in range(0, len(raw), dimensions)
    ]


def parse_pos(value: str) -> Coordinate:
    """ Parses a single gml:pos to a `Coordinate` structure.

        >>> parse_pos('12.34 56.7')
        (12.34, 56.7)
        >>> parse_pos('12.34 56.7 89.10')
        (12.34, 56.7, 89.1)
    """
    return tuple(float(v) for v in value.split())


def swap_coordinate_xy(coordinate: Coordinate) -> Coordinate:
    """ Swaps the X and Y coordinates of a given coordinate

        >>> swap_coordinate_xy((12.34, 56.7))
        (56.7, 12.34)
        >>> swap_coordinate_xy((12.34, 56.7, 89.10))
        (56.7, 12.34, 89.1)
    """
    return (coordinate[1], coordinate[0], *coordinate[2:])


def swap_coordinates_xy(coordinates: Coordinates) -> Coordinates:
    """ Swaps the X and Y coordinates of a given coordinates list

        >>> swap_coordinates_xy(
        ...     [(12.34, 56.7), (89.10, 11.12)]
        ... )
        [(56.7, 12.34), (11.12, 89.1)]
        >>> swap_coordinates_xy(
        ...     [(12.34, 56.7, 89.10), (11.12, 13.14, 15.16)]
        ... )
        [(56.7, 12.34, 89.1), (13.14, 11.12, 15.16)]
    """
    return [
        (coordinate[1], coordinate[0], *coordinate[2:])
        for coordinate in coordinates
    ]



# ------------------------------------------------------------------------------
#
# Project: pygml <https://github.com/geopython/pygml>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from typing import List

from lxml import etree

from .types import GeomDict
from .v3_common import (
    GML3Encoder, GML3Parser, parse_envelope, parse_point, parse_multi_point,
    parse_linestring_or_linear_ring,
    parse_multi_curve, parse_polygon, parse_multi_surface,
    parse_multi_geometry
)

NAMESPACE = 'http://www.opengis.net/gml/3.2'
NSMAP = {'gml': NAMESPACE}


Element = etree._Element
Elements = List[Element]

# set up a parser
GML32_PARSER = GML3Parser([NAMESPACE], NSMAP, {
    'Point': parse_point,
    'MultiPoint': parse_multi_point,
    'LineString': parse_linestring_or_linear_ring,
    'MultiCurve': parse_multi_curve,
    'Polygon': parse_polygon,
    'Envelope': parse_envelope,
    'MultiSurface': parse_multi_surface,
    'MultiGeometry': parse_multi_geometry,
})


def parse_v32(element: Element) -> GeomDict:
    """ Main parsing function for GML 3.2 XML structures.

        The following XML tags can be parsed to their respective GeoJSON
        counterpart:

          - gml:Point -> Point
          - gml:MultiPoint -> MultiPoint
          - gml:LineString -> LineString
          - gml:MultiCurve (with only gml:LineString curve members)
            -> MultiLineString
          - gml:Polygon -> Polygon
          - gml:MultiSurface (with only gml:Polygon surface members)
            -> MultiPolygon
          - gml:MultiGeometry (with any of the aforementioned types as
            geometry members) -> GeometryCollection

        The SRS of the geometry is determined and the coordinates are
        flipped to XY order in GeoJSON when they are in YX order in GML.

        Returns:
            the parsed GeoJSON geometry as a dict. Contains a 'type'
            field, a 'coordinates' field and potentially a 'crs' field
            when the geometries SRS could be determined. This field
            follows the structure laid out in the
            `draft for GeoJSON <https://gist.github.com/sgillies/1233327>`_.
    """

    return GML32_PARSER.parse(element)


GML32_ENCODER = GML3Encoder(NAMESPACE, NSMAP, True)


def encode_v32(geometry: GeomDict, identifier: str) -> Element:
    """ Encodes the given GeoJSON dict to its most simple GML 3.2
        representation. As in GML 3.2 the gml:id attribute is mandatory,
        the identifier must be passed as well.

        In preparation of the encoding, the coordinates may have to be
        swapped from XY order to YX order, depending on the used CRS.
        This includes the case when no CRS is specified, as this means
        the default WGS84 in GeoJSON, which in turn uses
        latitude/longitude ordering GML.

        This function returns an ``lxml.etree._Element`` which can be
        altered or serialized.

        >>> from pygml.v32 import encode_v32
        >>> from lxml import etree
        >>> tree = encode_v32({
        ...     'type': 'Point',
        ...     'coordinates': (1.0, 1.0)
        ... }, 'ID')
        >>> print(etree.tostring(tree, pretty_print=True).decode())
        <gml:Point xmlns:gml="http://www.opengis.net/gml/3.2"
            srsName="urn:ogc:def:crs:OGC::CRS84" gml:id="ID">
          <gml:pos>1.0 1.0</gml:pos>
        </gml:Point>
    """

    return GML32_ENCODER.encode(geometry, identifier)
